from typing import Optional

from battlenet_client import utils


class LeagueData:
    @staticmethod
    def league_data(
        client,
        region_tag: str,
        season_id: str,
        queue_id: str,
        team_type: str,
        league_id: str,
        locale: Optional[str] = None,
    ):
        """Returns data for the specified season, queue, team, and league.

        queueId: the standard available queueIds are: 1=WoL 1v1, 2=WoL 2v2, 3=WoL 3v3, 4=WoL 4v4, 101=HotS 1v1,
            102=HotS 2v2, 103=HotS 3v3, 104=HotS 4v4, 201=LotV 1v1, 202=LotV 2v2, 203=LotV 3v3, 204=LotV 4v4,
            206=LotV Archon. Note that other available queues may not be listed here.

        teamType: there are two available teamTypes: 0=arranged, 1=random.

        leagueId: available leagueIds are: 0=Bronze, 1=Silver, 2=Gold, 3=Platinum, 4=Diamond, 5=Master, 6=Grandmaster.

        Args:
            client (obj: oauth): OpenID/OAuth instance
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            season_id (str): The season ID of the data to retrieve.
            queue_id (str): The queue ID of the data to retrieve.
            team_type (str): The team type of the data to retrieve.
            league_id (str): The league ID of the data to retrieve.

        Returns:
            dict: dict containing league data for specified season, queue, team, and league.
        """
        uri = f"{utils.api_host(region_tag)}/data/sc2/league/{season_id}/{queue_id}/{team_type}/{league_id}"
        params = {"locale": utils.localize(locale)}
        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)



"""Defines exceptions related to the Diablo III API wrappers"""


class D3Error(Exception):
    """Base Error class for Diablo III Client"""

    pass


class D3ClientError(D3Error):
    """Error raised if there is a mismatch with the client and API endpoints"""

    pass



"""Defines exceptions related to the Starcraft 2 API wrappers

Classes:
    SC2Error
    SC2ClientError
    SC2RegionError

Disclaimer:
    All rights reserved, Blizzard is the intellectual property owner of Starcraft 2 and any data
    retrieved from this API.

"""


class SC2Error(Exception):
    """Base Exception class for Starcraft 2"""

    pass


class SC2ClientError(SC2Error):
    """Exception when the client is not the correct one"""

    pass


class SC2RegionError(SC2Error):
    """Exception for when there is a regional data conflict"""

    pass



"""This module contains the classes for accessing profile related APIs
"""
from urllib.parse import quote
from typing import Optional


from battlenet_client import utils
from .exceptions import WoWReleaseError


class Account:
    @staticmethod
    def account_profile_summary(
        client,
        region_tag: str,
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Accesses a summary of the account

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request

        Returns:
            dict: JSON decoded data that contains the profile summary
        """
        uri = f"{utils.api_host(region_tag)}/profile/user/wow"
        params = {
            "locale": utils.localize(locale),
            "namespace": utils.namespace("profile", release, region_tag),
        }
        return client.get(uri, params=params)

    @staticmethod
    def protected_character_profile_summary(
        client,
        region_tag: str,
        realm_id: int,
        character_id: int,
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Accesses a summary of protected account information for the
        character identified by :realm_id: and :character_id:

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            realm_id (int): the ID for the character's realm
            character_id (int): the ID of character

        Returns:
            dict: JSON decoded data that contains the protected character
                profile summary
        """
        uri = f"{utils.api_host(region_tag)}/profile/user/wow/protected-character/{realm_id}-{character_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": utils.namespace("profile", release, region_tag),
        }
        return client.get(uri, params=params)

    @staticmethod
    def account_collections(
        client,
        region_tag: str,
        *,
        locale: Optional[str] = None,
        category: Optional[str] = "",
        release: Optional[str] = "retail",
    ):
        """Access the collection of battle pets and/or mounts of an account as
        provided by :category:

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            category (str): 'pets' to retrieve the pet collections, and
                'mounts' to retrieve the mount collection of the account or
                None for both pets and mounts

        Returns:
            dict: JSON decoded data for the index/individual collections
        """
        uri = f"{utils.api_host(region_tag)}/profile/user/wow/collections/{category}"

        params = {
            "locale": utils.localize(locale),
            "namespace": utils.namespace("profile", release, region_tag),
        }
        return client.get(uri, params=params)


class Character:
    @staticmethod
    def achievement_summary(
        client,
        region_tag: str,
        realm_name: str,
        character_name: str,
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Accesses the achievement summary of the requested character
        identified by :character_name: on realm :realm_name:

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            realm_name (str): the slug for the character's realm
            character_name (str): name of character

        Returns:
            dict: JSON decoded data that contains requested achievement summary
        """
        uri = f"{utils.api_host(region_tag)}/profile/wow/character/"
        uri += (
            f"{utils.slugify(realm_name)}/{utils.slugify(character_name)}/achievements"
        )

        params = {
            "locale": utils.localize(locale),
            "namespace": utils.namespace("profile", release, region_tag),
        }
        return client.get(uri, params=params)

    @staticmethod
    def achievement_statistics(
        client,
        region_tag: str,
        realm_name: str,
        character_name: str,
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Accesses the achievement statistics for the requested character
        identified by :character_name: on realm :realm_name:

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            realm_name (str): the slug for the character's realm
            character_name (str): name of character

        Returns:
            dict: JSON decoded data that contains the requested achievement statistics
        """
        uri = f"{utils.api_host(region_tag)}/profile/wow/character/"
        uri += f"{utils.slugify(realm_name)}/{utils.slugify(character_name)}/achievements/statistics"
        params = {
            "locale": utils.localize(locale),
            "namespace": utils.namespace("profile", release, region_tag),
        }
        return client.get(uri, params=params)

    @staticmethod
    def appearance_summary(
        client,
        region_tag: str,
        realm_name: str,
        character_name: str,
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Accesses the appearance summary for the requested character
        identified by :character_name: on realm :realm_name:

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            realm_name (str): the slug for the character's realm
            character_name (str): name of character

        Returns:
            dict: JSON decoded data that contains requested appearance summary
        """
        uri = f"{utils.api_host(region_tag)}/profile/wow/character/"
        uri += f"{utils.slugify(realm_name)}/{utils.slugify(character_name)}/appearance"
        params = {
            "locale": utils.localize(locale),
            "namespace": utils.namespace("profile", release, region_tag),
        }
        return client.get(uri, params=params)

    @staticmethod
    def collections(
        client,
        region_tag: str,
        realm_name: str,
        character_name: str,
        *,
        locale: Optional[str] = None,
        category: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Accesses the battle pet and/or mount collections for the requested
        character identified by :character_name: on realm :realm_name: of the
        given :category:.

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            realm_name (str): the slug for the character's realm
            character_name (str): name of character
            category (str): category to retrieve. options are pets or mounts, or None (default).  None will
                provide both

        Returns:
            dict: JSON decoded data that contains requested collection
        """
        uri = f"{utils.api_host(region_tag)}/profile/wow/character/"
        uri += (
            f"{utils.slugify(realm_name)}/{utils.slugify(character_name)}/collections"
        )

        if category:
            if category.lower() not in ("pets", "mounts"):
                raise ValueError("Category needs to pets or mounts")
            uri += f"/{utils.slugify(category)}"

        params = {
            "locale": utils.localize(locale),
            "namespace": utils.namespace("profile", release, region_tag),
        }
        return client.get(uri, params=params)

    @staticmethod
    def encounters(
        client,
        region_tag: str,
        realm_name: str,
        character_name: str,
        *,
        locale: Optional[str] = None,
        category: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Accesses the encounters for the requested character identified by
        :character_name: on realm :realm_name:.  The encounters can be limited
        by :category:

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            realm_name (str): the slug for the character's realm
            character_name (str): name of character
            category (str): category to retrieve.  options are 'dungeons',
                'raids', or None (default).  None will access both dungeons and raids

        Returns:
            dict: JSON decoded dict that contains requested encounter data or index
        """
        uri = f"{utils.api_host(region_tag)}/profile/wow/character/"
        uri += f"{utils.slugify(realm_name)}/{utils.slugify(character_name)}/encounters"

        if category:
            if category.lower() not in ("dungeons", "raids"):
                raise ValueError("Available Categories: None, dungeons and raids")
            uri += f"/{utils.slugify(category)}"

        params = {
            "locale": utils.localize(locale),
            "namespace": utils.namespace("profile", release, region_tag),
        }
        return client.get(uri, params=params)

    @staticmethod
    def equipment_summary(
        client,
        region_tag: str,
        realm_name: str,
        character_name: str,
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Accesses the equipped items of the requested character identified by
        :character_name: on realm :realm_name:

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            realm_name (str): the slug for the character's realm
            character_name (str): name of character

        Returns:
            dict: JSON decoded dict that contains requested equipment summary
        """
        uri = f"{utils.api_host(region_tag)}/profile/wow/character/"
        uri += f"{utils.slugify(realm_name)}/{utils.slugify(character_name)}/equipment"

        params = {
            "locale": utils.localize(locale),
            "namespace": utils.namespace("profile", release, region_tag),
        }
        return client.get(uri, params=params)

    @staticmethod
    def hunter_pets_summary(
        client,
        region_tag: str,
        realm_name: str,
        character_name: str,
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Access the list of hunter pets of the requested character identified
        by :character_name: on realm :realm_name:

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            realm_name (str): the slug for the character's realm
            character_name (str): name of character

        Returns:
            dict: JSON decoded data of the hunter's pets
        """
        uri = f"{utils.api_host(region_tag)}/profile/wow/character/"
        uri += (
            f"{utils.slugify(realm_name)}/{utils.slugify(character_name)}/hunter-pets"
        )

        params = {
            "locale": utils.localize(locale),
            "namespace": utils.namespace("profile", release, region_tag),
        }
        return client.get(uri, params=params)

    @staticmethod
    def media_summary(
        client,
        region_tag: str,
        realm_name: str,
        character_name: str,
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Accesses the media assets, such as avatar render, of the requested
        character identified by :character_name: on realm :realm_name:

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            realm_name (str): the slug for the character's realm
            character_name (str): name of character

        Returns:
            dict: JSON decoded data that contains requested media assets
        """
        uri = f"{utils.api_host(region_tag)}/profile/wow/character/"
        uri += f"{utils.slugify(realm_name)}/{utils.slugify(character_name)}/character-media"

        params = {
            "locale": utils.localize(locale),
            "namespace": utils.namespace("profile", release, region_tag),
        }
        return client.get(uri, params=params)

    @staticmethod
    def mythic_keystone(
        client,
        region_tag: str,
        realm_name: str,
        character_name: str,
        *,
        locale: Optional[str] = None,
        season_id: Optional[int] = None,
        release: Optional[str] = "retail",
    ):
        """Accesses the mythic keystone (M+ or Mythic+) information of the
        requested character identified by :character_name: on realm
        :realm_name:

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            realm_name (str): the slug for the character's realm
            character_name (str): name of character
            season_id (int or None): season id or None (default).  None
                accesses the list of seasons for the current expansion

        Returns:
            dict: JSON decoded data of requested mythic keystone details
        """
        uri = f"{utils.api_host(region_tag)}/profile/wow/character/"
        uri += f"{utils.slugify(realm_name)}/{utils.slugify(character_name)}/mythic-keystone-profile"

        if season_id:
            uri += f"/season/{season_id}"

        params = {
            "locale": utils.localize(locale),
            "namespace": utils.namespace("profile", release, region_tag),
        }
        return client.get(uri, params=params)

    @staticmethod
    def professions_summary(
        client,
        region_tag: str,
        realm_name: str,
        character_name: str,
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Accesses the profession information of the requested character
        identified by :character_name: on realm :realm_name:

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            realm_name (str): the slug for the character's realm
            character_name (str): name of character

        Returns:
            dict: JSON decoded data of the character's professions
        """
        uri = f"{utils.api_host(region_tag)}/profile/wow/character/"
        uri += (
            f"{utils.slugify(realm_name)}/{utils.slugify(character_name)}/professions"
        )

        params = {
            "locale": utils.localize(locale),
            "namespace": utils.namespace("profile", release, region_tag),
        }
        return client.get(uri, params=params)

    @staticmethod
    def profile(
        client,
        region_tag: str,
        realm_name: str,
        character_name: str,
        *,
        locale: Optional[str] = None,
        status: bool = False,
        release: Optional[str] = "retail",
    ):
        """Access the profile status of the requested character identified by
        :character_name: on realm :realm_name:

        When requesting the character profile status, a note from Blizzard:

        Returns the status and a unique ID for a character. A client should
        delete information about a character from their application if
        any of the following conditions occur:

            1) an HTTP 404 Not Found error is returned
            2) the is_valid value is false
            3) the returned character ID doesn't match the previously recorded
                value for the character

        The following example illustrates how to use this endpoint:
            1) A client requests and stores information about a character, including
            its unique character ID and the timestamp of the request.
            2) After 30 days, the client makes a request to the status endpoint to
            verify if the character information is still valid.
            3) If character cannot be found, is not valid, or the characters IDs do
            not match, the client removes the information from their application.
            4) If the character is valid and the character IDs match, the client retains
            the data for another 30 days.

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            realm_name (str): the slug for the character's realm
            character_name (str): name of character
            status (bool): flag to request a profile summary (False default) or status (True)

        Returns:
            dict: JSON decoded data of character profile summary
        """
        uri = f"{utils.api_host(region_tag)}/profile/wow/character/"
        uri += f"{utils.slugify(realm_name)}/{quote(character_name.lower())}"

        if status:
            uri += "/status"

        params = {
            "locale": utils.localize(locale),
            "namespace": utils.namespace("profile", release, region_tag),
        }
        return client.get(uri, params=params)

    @staticmethod
    def pvp(
        client,
        region_tag: str,
        realm_name: str,
        character_name: str,
        *,
        locale: Optional[str] = None,
        pvp_bracket: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Accesses the Player versus Player (PvP) information of the requested
        character identified by :character_name: on realm :realm_name:

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            realm_name (str): the slug for the character's realm
            character_name (str): name of character
            pvp_bracket (str or None): '2v2', '3v3', 'rbg', None (default).
                None returns a summary of pvp activity

        Returns:
            dict: JSON decoded data of requested pvp details
        """
        uri = f"{utils.api_host(region_tag)}/profile/wow/character/"
        uri += f"{utils.slugify(realm_name)}/{utils.slugify(character_name)}"

        if pvp_bracket:
            uri += f"/pvp-bracket/{utils.slugify(pvp_bracket)}"
        else:
            uri += "/pvp-summary"

        params = {
            "locale": utils.localize(locale),
            "namespace": utils.namespace("profile", release, region_tag),
        }
        return client.get(uri, params=params)

    @staticmethod
    def quests(
        client,
        region_tag: str,
        realm_name: str,
        character_name: str,
        *,
        locale: Optional[str] = None,
        completed: Optional[bool] = False,
        release: Optional[str] = "retail",
    ):
        """Accesses the all or just completed quest information of the
        requested character identified by :character_name: on realm
        :realm_name:.

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            realm_name (str): the slug for the character's realm
            character_name (str): name of character
            completed (bool):  To show all quests (False), or to show only
                completed quests (True)

        Returns:
            dict: JSON decoded data of the completed or uncompleted quests
        """
        uri = f"{utils.api_host(region_tag)}/profile/wow/character/"
        uri += f"{utils.slugify(realm_name)}/{utils.slugify(character_name)}/quests"

        if completed:
            uri += f"/completed"

        params = {
            "locale": utils.localize(locale),
            "namespace": utils.namespace("profile", release, region_tag),
        }
        return client.get(uri, params=params)

    @staticmethod
    def reputations_summary(
        client,
        region_tag: str,
        realm_name: str,
        character_name: str,
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Accesses the reputation data of the requested character identified
        by :character_name: on realm :realm_name:

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            realm_name (str): the slug for the character's realm
            character_name (str): name of character

        Returns:
            dict: JSON decoded data of the character's reputations
        """
        uri = f"{utils.api_host(region_tag)}/profile/wow/character/"
        uri += (
            f"{utils.slugify(realm_name)}/{utils.slugify(character_name)}/reputations"
        )

        params = {
            "locale": utils.localize(locale),
            "namespace": utils.namespace("profile", release, region_tag),
        }
        return client.get(uri, params=params)

    @staticmethod
    def soulbinds(
        client,
        region_tag: str,
        realm_name: str,
        character_name: str,
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Accesses the available soulbinds of the requested character
        identified by :character_name: on realm :realm_name:

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            realm_name (str): the slug for the character's realm
            character_name (str): name of character

        Returns:
            dict: JSON decoded data of the character's soulbinds
        """
        uri = f"{utils.api_host(region_tag)}/profile/wow/character/"
        uri += f"{utils.slugify(realm_name)}/{utils.slugify(character_name)}/soulbinds"

        params = {
            "locale": utils.localize(locale),
            "namespace": utils.namespace("profile", release, region_tag),
        }
        return client.get(uri, params=params)

    @staticmethod
    def specializations_summary(
        client,
        region_tag: str,
        realm_name: str,
        character_name: str,
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Access the available specializations of the requested character
        identified by :character_name: on realm :realm_name:

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            realm_name (str): the slug for the character's realm
            character_name (str): name of character

        Returns:
            dict: JSON decoded data of the character's specializations
        """
        uri = f"{utils.api_host(region_tag)}/profile/wow/character/"
        uri += f"{utils.slugify(realm_name)}/{utils.slugify(character_name)}/specializations"

        params = {
            "locale": utils.localize(locale),
            "namespace": utils.namespace("profile", release, region_tag),
        }
        return client.get(uri, params=params)

    @staticmethod
    def statistics_summary(
        client,
        region_tag: str,
        realm_name: str,
        character_name: str,
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Accesses the statistics of the requested character identified by
        :character_name: on realm :realm_name:

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            realm_name (str): the slug for the character's realm
            character_name (str): name of character

        Returns:
            dict: JSON decoded data of the character's statistics
        """
        uri = f"{utils.api_host(region_tag)}/profile/wow/character/"
        uri += f"{utils.slugify(realm_name)}/{utils.slugify(character_name)}/statistics"

        params = {
            "locale": utils.localize(locale),
            "namespace": utils.namespace("profile", release, region_tag),
        }
        return client.get(uri, params=params)

    @staticmethod
    def title_summary(
        client,
        region_tag: str,
        realm_name: str,
        character_name: str,
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Accesses the list of titles earned by the requested character
        identified by :character_name: on realm :realm_name:

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            realm_name (str): the slug for the character's realm
            character_name (str): name of character

        Returns:
            dict: JSON decoded data the titles the player has earned
        """
        uri = f"{utils.api_host(region_tag)}/profile/wow/character/"
        uri += f"{utils.slugify(realm_name)}/{utils.slugify(character_name)}/titles"

        params = {
            "locale": utils.localize(locale),
            "namespace": utils.namespace("profile", release, region_tag),
        }
        return client.get(uri, params=params)


class Guild:
    @staticmethod
    def guild(
        client,
        region_tag: str,
        realm_name: str,
        guild_name: str,
        *,
        category: Optional[str] = None,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns a single guild by its name and realm.

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            realm_name (str): the name of the guild's realm
            guild_name (str): the name of the guild
            category (str): category of guild data to retrieve

        Returns:
            dict: json decoded data of the guild
        """
        if release != "retail":
            raise WoWReleaseError(
                f"{release.title()} does not support the Guild Data API"
            )

        uri = f"{utils.api_host(region_tag)}/data/wow/guild/"
        uri += f"{utils.slugify(realm_name)}/{utils.slugify(guild_name)}"

        if category and category in ("activity", "achievements", "roster"):
            uri += f"/{category}"

        params = {
            "locale": utils.localize(locale),
            "namespace": utils.namespace("profile", release, region_tag),
        }
        return client.get(uri, params=params)



"""This module contains the classes for accessing game data related utils.apis"""
from typing import Optional, Any, Dict, Union

from .exceptions import WoWReleaseError
from battlenet_client import utils

from battlenet_client.wow.utils import namespace


class Achievement:

    @staticmethod
    def achievement_category(
        client,
        region_tag: str,
        *,
        release: Optional[str] = "retail",
        category_id: Optional[int] = "index",
        locale: Optional[str] = None,
    ):
        """Accesses a list achievement categories or specific achievement
        category if :category_id: is provided

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            category_id (int, optional): the achievement's category ID or None (default).
                None will retrieve the entire list of achievement categories

        Returns:

            dict: json decoded data for the index/individual achievement categories
        """
        uri = (
            f"{utils.api_host(region_tag)}/data/wow/achievement-category/{category_id}"
        )
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace('static', release, region_tag),
        }
        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def achievement(
        client,
        region_tag: str,
        *,
        achievement_id: Optional[int] = "index",
        release: Optional[str] = "retail",
        locale: Optional[str] = None,
    ):
        """Returns an index of achievements, or a specific achievements

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            achievement_id (int, optional): the achievement ID or the word 'index'

        Returns:
            dict: json decoded data for the index/individual achievements
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/achievement/{achievement_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }
        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def achievement_media(
        client,
        region_tag: str,
        achievement_id: int,
        *,
        release: Optional[str] = "retail",
        locale: Optional[str] = None,
    ):
        """Returns media for an achievement's icon.

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            achievement_id (int): the achievement ID or the word 'index'

        Returns:
            dict: json decoded media data for the achievement
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/media/{achievement_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)


class Auction:
    @staticmethod
    def auction(
        client,
        region_tag: str,
        connected_realm_id: int,
        *,
        release: Optional[str] = "retail",
        auction_house_id: Optional[int] = None,
        locale: Optional[str] = None,
    ):
        """Returns auction data.  With retail client, region_tag, it returns all the auctions for the given connected
        realm. For classic titles, the results can be either the entire list, or the individual auctions

        See the Connected Realm utils.api for information about retrieving a list of
        connected realm IDs.

        Auction house data updates at a set interval. The value was initially set
        at 1 hour; however, it might change over time without notice.

        Depending on the number of active auctions on the specified connected realm,
        the response from this game_data may be rather large, sometimes exceeding
        10 MB.

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            connected_realm_id (int): the id of the connected realm
            auction_house_id (int, optional): the ID of the auction house

        Returns:
            dict: json decoded data for the index/individual auction(s)

        Raises:
            WoWReleaseError: when an AH ID is used for the retail client
        Notes:
            Auction house functionality is not available for WoW 1.x (Vanilla Classic)
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/connected-realm/{connected_realm_id}/"
        if release == "retail" and auction_house_id is None:
            uri += "auctions"

        if release != "retail" and auction_house_id is None:
            uri += "auctions/index"

        if release != "retail" and auction_house_id is not None:
            uri += "auctions"

        if release == "retail" and auction_house_id is not None:
            raise WoWReleaseError("Auction House ID provided for retail")

        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)


class AzeriteEssence:
    @staticmethod
    def azerite_essence(
        client,
        region_tag: str,
        *,
        essence_id: Optional[int] = "index",
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of Azerite Essences, or a specific Azerite Essence

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            essence_id (int, optional): the Azerite essence ID or the word 'index'

        Returns:
            dict: json decoded data for the index/individual azerite essence(s)
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/azerite-essence/{essence_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def azerite_essence_search(
        client,
        region_tag: str,
        field_values: Dict[str, Any],
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Searches for azerite essences that match `field_values`

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            field_values (dict): search criteria, as key/value pairs

        Returns:
            dict: json decoded search results that match `field_values`
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/search/azerite-essence"

        # adding locale and namespace key/values pairs for a complete parameters list for the request
        field_values.update(
            {
                "locale": utils.localize(locale),
                "namespace": namespace("static", release, region_tag),
            }
        )

        try:
            return client.get(uri, params=field_values)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=field_values)

    @staticmethod
    def azerite_essence_media(
        client,
        region_tag: str,
        essence_id: int,
        *,
        release: Optional[str] = "retail",
        locale: Optional[str] = None,
    ):
        """Returns media data for an azerite essence.

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            essence_id (int): the azerite essence ID

        Returns:
            dict: json decoded media data for the azerite essence
        """
        uri = (
            f"{utils.api_host(region_tag)}/data/wow/media/azerite-essence/{essence_id}"
        )
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)


class ConnectedRealm:
    @staticmethod
    def connected_realm(
        client,
        region_tag: str,
        *,
        connected_realm_id: Optional[int] = "index",
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of connected realms, or a specific connected realm. Connected realm is a group of standard
        realms that act as one large realm

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            connected_realm_id (int, optional): the ID of the connected realm

        Returns:
            dict: json decoded data for the index/individual connected realms
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/connected-realm/{connected_realm_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def connected_realm_search(
        client,
        region_tag: str,
        field_values: Dict[str, Any],
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Searches the connected realm utils.api for connected realm(s) that match the criteria

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            field_values (dict): field/value pairs

        Returns:
            dict: json decoded search results that match `field_values`
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/search/connected_realm"

        # adding locale and namespace key/values pairs for a complete parameters list for the request
        field_values.update(
            {
                "locale": utils.localize(locale),
                "namespace": namespace("static", release, region_tag),
            }
        )

        try:
            return client.get(uri, params=field_values)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=field_values)


class Covenant:
    @staticmethod
    def covenant(
        client,
        region_tag: str,
        *,
        covenant_id: Optional[int] = "index",
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of covenants, or a specific covenant

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            covenant_id (int, optional): the ID of the covenant or the default 'index'

        Returns:
            dict: json decoded data for the index/individual covenant
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/covenant/{covenant_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def covenant_media(
        client,
        region_tag: str,
        covenant_id: int,
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns media for a covenant.

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            covenant_id (int, optional): the covenant ID

        Returns:
            dict: json decoded media data for the covenant
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/media/covenant{covenant_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def soulbind(
        client,
        region_tag: str,
        *,
        soulbind_id: Optional[int] = "index",
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of soulbinds, or a specific soulbind

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            soulbind_id (int, optional): the ID of the soulbind or the word of 'index'

        Returns:
            dict: json decoded data for the index/individual soulbind
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/covenant/soulbind/{soulbind_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def conduit(
        client,
        region_tag: str,
        *,
        conduit_id: Optional[int] = "index",
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of conduits, or a specific conduit

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            conduit_id (int, optional): the ID of the conduit or the word of 'index'

        Returns:
            dict: json decoded data for the index/individual conduit
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/covenant/conduit/{conduit_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)


class Creature:
    @staticmethod
    def creature_family(
        client,
        region_tag: str,
        *,
        family_id: Optional[int] = "index",
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of creature families, or a specific creature family

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            family_id (int, optional): the creature family ID or the default 'index'

        Returns:
            dict: json decoded data for the index/individual creature family/families
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/creature-family/{family_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def creature_type(
        client,
        region_tag: str,
        *,
        type_id: Optional[int] = "index",
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of creature types, or a specific creature type

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            type_id (int, optional): the creature type ID or the default 'index'

        Returns:
            dict: json decoded data for the index/individual creature type(s)
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/creature-type/{type_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def creature(
        client,
        region_tag: str,
        creature_id: int,
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of creatures, or a specific creature

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            creature_id (int, optional): the creature ID

        Returns:
            dict: json decoded data for the index/individual creature(s)
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/creature/{creature_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def creature_search(
        client,
        region_tag: str,
        field_values: Dict[str, Any],
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Searches the creature utils.api for creatures that match the criteria

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            field_values (dict): matching criteria in key/value pairs

        Returns:
            dict: json decoded search results that match `field_values`
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/search/creature"

        #  adding locale and namespace key/values pairs for a complete parameters list for the request
        field_values.update(
            {
                "locale": utils.localize(locale),
                "namespace": namespace("static", release, region_tag),
            }
        )

        try:
            return client.get(uri, params=field_values)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=field_values)

    @staticmethod
    def creature_display_media(
        client,
        region_tag: str,
        display_id: int,
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns media for a creature display.

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            display_id (int, optional): the creature display ID

        Returns:
            dict: json decoded media data for the creature display
        """
        uri = (
            f"{utils.api_host(region_tag)}/data/wow/media/creature-display/{display_id}"
        )
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def creature_family_media(
        client,
        region_tag: str,
        family_id: int,
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns media for a creature family.

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            family_id (int, optional): the creature family ID

        Returns:
            dict: json decoded media data for the creature family
        """
        uri = (
            f"{utils.api_host(region_tag)}/data/wow/media/creature-display/{family_id}"
        )
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)


class GuildCrest:
    @staticmethod
    def guild_crest_components_index(
        client,
        region_tag: str,
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of guild crest components.

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request

        Returns:
            dict: json decoded data for the index of guild crest components
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/guild-crest/index"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def guild_crest_border_media(
        client,
        region_tag: str,
        border_id: int,
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns media for a specific guild crest border.

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            border_id (int): the border ID

        Returns:
            dict: json decoded media data for the guild border
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/media/guild-crest/border/{border_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def guild_crest_emblem_media(
        client,
        region_tag: str,
        emblem_id: int,
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns media for a specific guild crest emblem.

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            emblem_id (int): the border ID

        Returns:
            dict: json decoded media data for the guild crest
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/media/guild-crest/emblem/{emblem_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)


class Item:
    @staticmethod
    def item_class(
        client,
        region_tag: str,
        *,
        class_id: Optional[int] = "index",
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of item classes, or a specific item class

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            class_id (int, optional): item class ID or the default 'index'

        Returns:
            dict: json decoded data for the index/individual item class(es)
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/item-class/{class_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def item_set(
        client,
        region_tag: str,
        *,
        set_id: Optional[int] = "index",
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of item sets, or a specific item set

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            set_id (int, optional): the item class ID or the default 'index'

        Returns:
            dict: json decoded data for the index/individual item set(s)
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/item-set/{set_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def item_subclass(
        client,
        region_tag: str,
        class_id: int,
        subclass_id: int,
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of item subclasses, or a specific item subclass

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            class_id (int): the item class ID
            subclass_id (int, optional): the item's subclass ID

        Returns:
            dict: json decoded data for the item's subclass
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/item-class/{class_id}/item-subclass/{subclass_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def item(
        client,
        region_tag: str,
        item_id: int,
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of items, or a specific item

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            item_id (int, optional): the item class ID

        Returns:
            dict: json decoded data for the individual item
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/item/{item_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def item_media(
        client,
        region_tag: str,
        item_id: int,
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns media for an item.

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            item_id (int): the creature family ID

        Returns:
            dict: json decoded media data for the item
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/media/item/{item_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def item_search(
        client,
        region_tag,
        field_values: Dict[str, Any],
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Searches the item utils.api for items that match the criteria

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            field_values (dict): search criteria as key/value pairs

        Returns:
             dict: json decoded search results that match `field_values`
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/search/item"

        #  adding locale and namespace key/values pairs to field_values to make a complete params list
        field_values.update(
            {
                "locale": utils.localize(locale),
                "namespace": namespace("static", release, region_tag),
            }
        )

        try:
            return client.get(uri, params=field_values)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=field_values)


class Journal:
    @staticmethod
    def journal_expansion(
        client,
        region_tag: str,
        *,
        expansion_id: Optional[int] = "index",
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of journal expansions, or a specific journal expansion

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            expansion_id (int, optional): the encounter ID or 'index'

        Returns:
            dict: json decoded data for the index/individual journal expansion(s)
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/journal-expansion/{expansion_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def journal_encounter(
        client,
        region_tag: str,
        *,
        encounter_id: Optional[int] = None,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of journal (boss) encounters, or a specific journal (boss) encounters

        Notes:
            This replaced the Boss endpoint of the community REST API

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            encounter_id (int, optional): the encounter ID or 'index'

        Returns:
            dict: json decoded data for the index/individual journal encounter(s)
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/journal-encounter/{encounter_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def journal_encounter_search(
        client,
        region_tag: str,
        field_values: Dict[str, Any],
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Searches for azerite essences that match `field_values`

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            field_values (dict): search criteria, as key/value pairs

        Returns:
            dict: json decoded search results that match `field_values`
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/search/journal-encounter"

        #  adding locale and namespace key/values pairs to field_values to make a complete params list
        field_values.update(
            {
                "locale": utils.localize(locale),
                "namespace": namespace("static", release, region_tag),
            }
        )

        try:
            return client.get(uri, params=field_values)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=field_values)

    @staticmethod
    def journal_instance(
        client,
        region_tag: str,
        *,
        instance_id: Optional[int] = "index",
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of journal instances (dungeons), or a specific journal instance (dungeon)

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            instance_id (int, optional): the encounter ID or 'index'

        Returns:
            dict: json decoded data for the index/individual journal instance(s)
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/journal-encounter/{instance_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def journal_instance_media(
        client,
        region_tag: str,
        instance_id: int,
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns media for an instance.

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            instance_id (int): the creature family ID

        Returns:
            dict: json decoded media data for the instance
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/media/journal-instance/{instance_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)


class Media:
    @staticmethod
    def media_search(
        client,
        region_tag: str,
        field_values: Dict[str, Any],
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Searches the media utils.api match the criteria

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            field_values (dict): fields and values for the search criteria

        Returns:
            dict: json decoded search results that match `field_values`
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/search/media"

        #  adding locale and namespace key/values pairs to field_values to make a complete params list
        field_values.update(
            {
                "locale": utils.localize(locale),
                "namespace": namespace("static", release, region_tag),
            }
        )

        try:
            return client.get(uri, params=field_values)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=field_values)


class ModifiedCrafting:
    @staticmethod
    def modified_crafting(
        client,
        region_tag: str,
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of modified crafting recipes

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request

        Returns:
            dict: json decoded data for the index of modified crafting
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/modified-crafting"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def modified_crafting_category(
        client,
        region_tag: str,
        *,
        category_id: Optional[int] = "index",
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of modified crafting category index, or a specific modified crafting category

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            category_id (int, optional): the encounter ID or 'index'

        Returns:
            dict: json decoded data for the index/individual modified crafting category/categories
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/modified-crafting/category/{category_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def modified_crafting_reagent_slot_type(
        client,
        region_tag: str,
        *,
        slot_type_id: Optional[int] = None,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of modified crafting reagent slot type, or a specific reagent slot type

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            slot_type_id (int, optional): the encounter ID or 'index'

        Returns:
            dict: json decoded data for the index/individual modified crafting reagent slot type(s)
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/modified-crafting/reagent-slot-type/{slot_type_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)


class Mount:
    @staticmethod
    def mount(
        client,
        region_tag: str,
        *,
        mount_id: Optional[int] = "index",
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of mounts, or a specific mount

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            mount_id (int, optional): the mount ID or 'index'

        Returns:
            dict: json decoded data for the index/individual mount(s)
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/mount/{mount_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def mount_search(
        client,
        region_tag: str,
        field_values: Dict[str, Any],
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Searches the mount utils.api that match the criteria

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            field_values (dict): fields and values for the search criteria

        Returns:
            dict: json decoded search results that match `field_values`
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/search/mount"

        #  adding locale and namespace key/values pairs to field_values to make a complete params list
        field_values.update(
            {
                "locale": utils.localize(locale),
                "namespace": namespace("static", release, region_tag),
            }
        )

        try:
            return client.get(uri, params=field_values)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=field_values)


class MythicKeystone:
    @staticmethod
    def mythic_keystone_affix(
        client,
        region_tag: str,
        *,
        affix_id: Optional[int] = "index",
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of mythic keystone affixes, or a specific mythic keystone affix

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            affix_id (int, optional): the affix's ID or the word 'index'

        Returns:
            dict: json decoded data for the index/individual mythic keystone affix(es)
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/keystone-affix/{affix_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def mythic_keystone_affix_media(
        client,
        region_tag: str,
        affix_id: int,
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns media for a mythic keystone affix.

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            affix_id (int): the affix's ID

        Returns:
            dict: json decoded media data for the mythic keystone affix(es)
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/media/keystone-affix/{affix_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def mythic_keystone_dungeon(
        client,
        region_tag: str,
        *,
        dungeon_id: Optional[int] = "index",
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of mythic keystone dungeons, or a specific mythic keystone dungeon

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            dungeon_id (int, optional): the dungeon's ID or 'index'

        Returns:
            dict: json decoded data for the index/individual mythic keystone dungeon(s)
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/mythic-keystone/dungeon/{dungeon_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def mythic_keystone_index(
        client,
        region_tag: str,
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of links to other documents related to Mythic Keystone dungeons.

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request

        Returns:
            dict: json decoded data for the index of the mythic keystone dungeon documents
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/mythic-keystone/index"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def mythic_keystone_period(
        client,
        region_tag: str,
        *,
        period_id: Optional[int] = "index",
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of mythic keystone periods, or a specific mythic keystone period

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            period_id (int, optional): the keystone's period ID or the word 'index'

        Returns:
            dict: json decoded data for the index/individual for mythic keystone period(s)
        """
        uri = (
            f"{utils.api_host(region_tag)}/data/wow/mythic-keystone/period/{period_id}"
        )
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def mythic_keystone_season(
        client,
        region_tag: str,
        *,
        season_id: Optional[int] = "index",
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of mythic keystone seasons, or a specific mythic keystone seasons

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            season_id (int, optional): the keystone's season ID or the word 'index'

        Returns:
            dict: json decoded data for the index/individual mythic keystone season(s)
        """
        uri = (
            f"{utils.api_host(region_tag)}/data/wow/mythic-keystone/season/{season_id}"
        )
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def mythic_keystone_leader_board(
        client,
        region_tag: str,
        connected_realm_id: int,
        *,
        dungeon_id: Optional[int] = None,
        period_id: Optional[int] = None,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of mythic keystone leader boards, or a specific mythic keystone leader board

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            connected_realm_id (int): the connected realm's id
            dungeon_id (int, optional): the particular dungeon's ID or the word 'index'
            period_id (int, optional): the particular period to search or None when looking for the index

        Returns:
            dict: json decoded data for the index/individual mythic keystone leaderboard(s)
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/connected-realm/{connected_realm_id}/mythic-leaderboard/"

        if dungeon_id and period_id:
            uri += "{dungeon_id}/period/{period_id}"
        else:
            uri += "index"

        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)


class MythicRaid:
    @staticmethod
    def mythic_raid_leaderboard(
        client,
        region_tag: str,
        raid_name: str,
        faction: str,
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of mythic keystone affixes, or a specific mythic keystone affix

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            raid_name (str): name of the raid
            faction (str): horde or alliance, defaults to alliance

        Returns:
            dict: json decoded data for the index/individual mythic raid
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/leaderboard/hall-of-fame/"
        uri += f"{utils.slugify(raid_name)}/{utils.slugify(faction)}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)


class Pet:
    @staticmethod
    def pet(
        client,
        region_tag: str,
        *,
        pet_id: Optional[int] = "index",
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of pets, or the data about the specified pet

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            pet_id (int, optional): the pet ID or the word 'index'

        Returns:
            dict: json decoded data for the index/individual pet(s)
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/pet/{pet_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def pet_media(
        client,
        region_tag: str,
        pet_id: int,
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns media for a pet

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            pet_id (int): the azerite pet ID

        Returns:
            dict: json decoded media data for the for pet
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/media/pet/{pet_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def pet_ability(
        client,
        region_tag: str,
        *,
        pet_ability_id: Optional[int] = "index",
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of pets, or the data about the specified pet

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            pet_ability_id (int, optional): the pet ID or the word 'index'

        Returns:
            dict: json decoded data for the index/individual pet ability/abilities
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/pet-ability/{pet_ability_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def pet_ability_media(
        client,
        region_tag: str,
        ability_id: int,
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns media for an azerite ability.

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            ability_id (int): the azerite ability ID

        Returns:
            dict: json decoded media data for the pet ability
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/media/pet-ability/{ability_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)


class PlayableClass:
    @staticmethod
    def playable_class(
        client,
        region_tag: str,
        *,
        class_id: Optional[int] = "index",
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of playable classes, or a specific playable class

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            class_id (int, optional): the class ID or the word 'index'

        Returns:
            dict: json decoded data for the index/individual playable class(es)
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/playable-class/{class_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def playable_class_media(
        client,
        region_tag: str,
        class_id: int,
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns media for a playable class by ID.

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            class_id (int ): class id

        Returns:
            dict: json decoded media data for the playable class(es)
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/media/playable-class/{class_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def pvp_talent_slots(
        client,
        region_tag: str,
        class_id: int,
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns the PvP talent slots for a playable class by ID.

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            class_id (int): class id

        Returns:
            dict: json decoded data for the index of PvP Talent slots
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/playable-class/{class_id}/pvp-talent-slots"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)


class PlayableRace:
    @staticmethod
    def playable_race(
        client,
        region_tag: str,
        *,
        race_id: Optional[int] = "index",
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of playable races, or a specific playable race

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            race_id (int, optional): the playable race's ID or the word 'index'

        Returns:
            dict: json decoded data for the index/individual playable race(s)
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/playable-race/{race_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)


class PlayableSpec:
    @staticmethod
    def playable_spec(
        client,
        region_tag: str,
        *,
        spec_id: Optional[int] = "index",
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):

        """Returns an index of playable specialization, or a specific playable specialization

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            spec_id (int, optional): the playable specialization's ID or the word 'index'

        Returns:
            dict: json decoded data for the index/individual playable specialization(s)
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/playable-specialization/{spec_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def playable_spec_media(
        client,
        region_tag: str,
        spec_id: int,
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns media for a playable specialization by ID.

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            spec_id (int): specialization id

        Returns:
            dict: json decoded media data for the playable specialization
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/media/playable-specialization/{spec_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)


class PowerType:
    @staticmethod
    def power_type(
        client,
        region_tag: str,
        *,
        power_id: Optional[int] = "index",
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of power types, or a specific power type

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            power_id (int, optional): the power type's ID or the word 'index'

        Returns:
            dict: json decoded data for the index/individual power types
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/power-type/{power_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)


class Profession:
    @staticmethod
    def profession(
        client,
        region_tag: str,
        *,
        profession_id: Optional[int] = "index",
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of achievements, or a specific achievements

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            profession_id (int, optional): the profession ID or the word 'index'

        Returns:
            dict: json decoded dict for the profession or the index of the achievements
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/profession/{profession_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def profession_media(
        client,
        region_tag: str,
        profession_id: int,
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns media for a creature display.

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            profession_id (str):  profession ID

        Returns:
            dict: the media assets for the given creature display ID
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/media/profession/{profession_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def profession_skill_tier(
        client,
        region_tag: str,
        profession_id: int,
        skill_tier_id: int,
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of achievements, or a specific achievements

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            profession_id (int): the profession ID
            skill_tier_id (int): the skill tier ID

        Returns:
            dict: json decoded dict for the profession or the index of the achievements
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/profession/{profession_id}/skill-tier/{skill_tier_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def recipe(
        client,
        region_tag: str,
        recipe_id: int,
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of achievements, or a specific achievements

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            recipe_id (str): the recipe ID

        Returns:
            dict: json decoded dict for the profession or the index of the achievements
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/recipe/{recipe_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def recipe_media(
        client,
        region_tag: str,
        recipe_id: int,
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns media for a creature display.

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            recipe_id (int): the profession ID

        Returns:
            dict: the media assets for the given creature display ID
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/media/recipe/{recipe_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)


class PVP:
    @staticmethod
    def pvp_season(
        client,
        region_tag: str,
        *,
        season_id: Optional[int] = "index",
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of pvp seasons, or a specific pvp season

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            season_id (int, optional): the power type's ID or the word 'index'

        Returns:
            dict: json decoded data for the index/individual PvP season(s)
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/pvp-season/{season_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def pvp_leader_board(
        client,
        region_tag,
        season_id: int,
        *,
        pvp_bracket: Optional[str] = "index",
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of pvp leader boards, or a specific pvp leader board

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            season_id (int): pvp season's ID
            pvp_bracket (int, optional): the PvP bracket to view ('2v2', '3v3', '5v5', 'rbg') or the word 'index'

        Returns:
            dict: json decoded data for the index of the PvP leader board
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/pvp-season/{season_id}/pvp-leaderboard/{pvp_bracket}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def pvp_rewards_index(
        client,
        region_tag: str,
        season_id: int,
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of pvp rewards, or a specific pvp reward

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            season_id (int): the season ID for the rewards or the word 'index'

        Returns:
            dict: json decoded data for the index of PvP rewards
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/pvp-season/{season_id}/pvp-reward/index"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def pvp_tier(
        client,
        region_tag: str,
        *,
        tier_id: Optional[int] = "index",
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of pvp tier, or a specific pvp tier

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            tier_id (int, optional): the pvp tier ID or the default 'index'

        Returns:
            dict: the index or data for the pvp tier
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/pvp-tier/{tier_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def pvp_tier_media(
        client,
        region_tag,
        tier_id: int,
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns media for a PvP tier by ID.

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            tier_id (int): pvp tier id

        Returns:
            dict: json decoded media data for the PvP tier
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/media/pvp-tier/{tier_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)


class Quest:
    @staticmethod
    def quest(
        client,
        region_tag: str,
        *,
        quest_id: Optional[int] = "index",
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of quests, or a specific quest

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            quest_id (int, optional): the quest ID or the word 'index'

        Returns:
            dict: json decoded data for the index/individual quest(s)
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/quest/{quest_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def quest_category(
        client,
        region_tag: str,
        *,
        quest_category_id: Optional[int] = "index",
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of quest categories, or a specific quest category

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            quest_category_id (int, optional): the quest category ID or the word 'index'

        Returns:
            dict: json decoded data for the index/individual quest category/categories
        """
        uri = (
            f"{utils.api_host(region_tag)}/data/wow/quest/category/{quest_category_id}"
        )
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def quest_area(
        client,
        region_tag: str,
        *,
        quest_area_id: Optional[int] = "index",
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of quest areas, or a specific quest area

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            quest_area_id (int, optional): the quest area ID or the word 'index'

        Returns:
            dict: json decoded data for the index/individual quest area(s)
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/quest/area/{quest_area_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def quest_type(
        client,
        region_tag: str,
        *,
        quest_type_id: Optional[int] = "index",
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of quest types, or a specific quest type

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            quest_type_id (int, optional): the quest type ID or the word 'index'

        Returns:
            dict: json decoded data for the index/individual quest type(s)
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/quest/type/{quest_type_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)


class Realm:
    @staticmethod
    def realm(
        client,
        region_tag: str,
        *,
        realm_slug: Optional[Union[str, int]] = "index",
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of realms, or a specific realm

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            realm_slug (str/int, optional): the pvp tier ID or the word 'index'

        Returns:
            dict: json decoded data for the index/individual realm(s)
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/realm/{realm_slug}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def realm_search(
        client,
        region_tag,
        field_values: Dict[str, Any],
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):

        """Searches the creature utils.api for connected realm(s) that match the criteria

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            field_values (dict): search criteria, as key/value pairs

        Returns:
            dict: json decoded search results that match `field_values`
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/search/realm"

        #  adding locale and namespace key/values pairs to field_values to make a complete params list
        field_values.update(
            {
                "locale": utils.localize(locale),
                "namespace": namespace("static", release, region_tag),
            }
        )

        try:
            return client.get(uri, params=field_values)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=field_values)


class Region:
    @staticmethod
    def region(
        client,
        region_tag: str,
        *,
        region_req: Optional[int] = "index",
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of region_tags, or a specific region_tag

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            region_req (int, optional): the region_tag ID or the word 'index'

        Returns:
            dict: json decoded data for the index/individual region_tag(s)
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/region/{region_req}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)


class Reputation:
    @staticmethod
    def reputation_faction(
        client,
        region_tag: str,
        *,
        faction_id: Optional[int] = "index",
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):

        """Returns an index of reputation factions, or a specific reputation fa

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            faction_id (int, optional): the slug or ID of the region_tag requested

        Returns:
            dict: json decoded data for the index/individual reputation faction(s)
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/reputation-faction/{faction_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def reputation_tier(
        client,
        region_tag: str,
        *,
        tier_id: Optional[int] = "index",
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of reputation factions, or a specific reputation fa

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            tier_id (int, optional): the slug or ID of the region_tag requested

        Returns:
            dict: json decoded data for the index/individual reputation tier(s)
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/reputation-tiers/{tier_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)


class Spell:
    @staticmethod
    def spell(
        client,
        region_tag: str,
        spell_id: int,
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of spells, or a specific spell

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            spell_id (int): the slug or ID of the region_tag requested

        Returns:
            dict: json decoded data for the index/individual spell(s)
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/spell/{spell_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def spell_media(
        client,
        region_tag,
        spell_id: int,
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns media for a spell by ID.

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            spell_id (int): pvp tier id

        Returns:
            dict: json decoded media data for the spell
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/media/spell/{spell_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def spell_search(
        client,
        region_tag: str,
        field_values: Dict[str, Any] = None,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Searches the creature utils.api for items that match the criteria

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            field_values (dict): search criteria, as key/value pairs

        Returns:
            dict: json decoded search results that match `field_values`
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/search/spell"

        #  adding locale and namespace key/values pairs to field_values to make a complete params list
        field_values.update(
            {
                "locale": utils.localize(locale),
                "namespace": namespace("static", release, region_tag),
            }
        )

        try:
            return client.get(uri, params=field_values)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=field_values)


class Talent:
    @staticmethod
    def talent(
        client,
        region_tag: str,
        *,
        talent_id: Optional[int] = "index",
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of spells, or a specific spell

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            talent_id (int, optional): the slug or ID of the region_tag requested

        Returns:
            dict: json decoded data for the index/individual talent(s)
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/talent/{talent_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def pvp_talent(
        client,
        region_tag: str,
        *,
        pvp_talent_id: Optional[int] = "index",
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):

        """Returns an index of spells, or a specific spell

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            pvp_talent_id (int, optional): the slug or ID of the region_tag requested

        Returns:
            dict: json decoded data for the talent index or individual talent
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/pvp-talent/{pvp_talent_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)


class TechTalent:
    @staticmethod
    def tech_talent_tree(
        client,
        region_tag: str,
        *,
        tree_id: Optional[int] = "index",
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of tech talent trees or a tech talent tree by ID

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            tree_id (int, optional): the slug or ID of the region_tag requested

        Returns:
            dict: json decoded data for the index/individual tech talent tree(s)
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/tech-talent-tree/{tree_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def tech_talent(
        client,
        region_tag: str,
        *,
        talent_id: Optional[int] = "index",
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of tech talents or a tech talent by ID

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            talent_id (int, optional): the slug or ID of the region_tag requested

        Returns:
            dict: json decoded data for the index/individual tech talent(s)
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/tech-talent/{talent_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def tech_talent_media(
        client,
        region_tag: str,
        talent_id: int,
        *,
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns media for a spell by ID.

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            talent_id (int): pvp tier id

        Returns:
            dict: json decoded media data for the tech talent
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/media/tech-talent/{talent_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)


class Title:
    @staticmethod
    def title(
        client,
        region_tag: str,
        *,
        title_id: Optional[int] = "index",
        locale: Optional[str] = None,
        release: Optional[str] = "retail",
    ):
        """Returns an index of spells, or a specific spell

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            title_id (int, optional): the slug or ID of the region_tag requested

        Returns:
            dict: json decoded data for the index/individual title(s)
        """
        uri = f"{utils.api_host(region_tag)}/data/wow/title/{title_id}"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)


class WoWToken:
    @staticmethod
    def wow_token_index(
        client,
        region_tag,
        release: Optional[str] = "retail",
        locale: Optional[str] = None,
    ):
        """Returns the WoW Token index.

        Args:
            client (obj: oauth): OpenID/OAuth instance
            release (str): release of the game (ie classic1x, classic, retail)
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request

        Returns:
            dict: json decoded data for the index/individual wow token
        """
        if release in utils.WOW_CLASSICS and region_tag.lower() != "cn":
            raise WoWReleaseError(
                "WoW Token API only available on retail, and CN classic markets"
            )

        uri = f"{utils.api_host(region_tag)}/data/wow/token/index"
        params = {
            "locale": utils.localize(locale),
            "namespace": namespace("static", release, region_tag),
        }

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)



"""Defines the classes that handle the community APIs for Diablo III"""

from typing import Optional

from urllib.parse import quote

from battlenet_client import utils


class Community:
    """Defines the methods to communicate with the Community API"""

    @staticmethod
    def act(
        client,
        region_tag: str,
        *,
        act_id: Optional[int] = None,
        locale: Optional[str] = None,
    ):
        """Returns the list of acts, or the act by :act_id:

        Args:
            client (obj: oauth): OpenID/OAuth instance
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            act_id (int, optional): the act's ID to retrieve its data

        Returns:
            dict: the dict containing the list of acts or the details of the specified :act_id:
        """
        uri = f"{utils.api_host(region_tag)}/d3/data/act"

        if act_id:
            uri += f"/{act_id}"

        params = {"locale": utils.localize(locale)}

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_proctected_resource(uri, "GET", params=params)

    @staticmethod
    def artisan(
        client, region_tag: str, artisan_slug: str, locale: Optional[str] = None
    ):
        """Returns the artisan by the slug

        Args:
            client (obj: oauth): OpenID/OAuth instance
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            artisan_slug (str): the slug of the artisan

        Returns:
            dict: the dict containing data of the artisan
        """
        uri = f"{utils.api_host(region_tag)}/d3/data/artisan/{utils.slugify(artisan_slug)}"
        params = {"locale": utils.localize(locale)}

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_proctected_resource(uri, "GET", params=params)

    @staticmethod
    def recipe(
        client,
        region_tag: str,
        artisan_slug: str,
        recipe_slug: str,
        locale: Optional[str] = None,
    ):
        """Returns a single recipe by the by slug for the specified artisan

        Args:
            client (obj: oauth): OpenID/OAuth instance
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            artisan_slug (str): the slug of the artisan
            recipe_slug (str): the slug of the recipe

        Returns:
            dict: the dict containing for the recipe
        """
        uri = f"{utils.api_host(region_tag)}/d3/data/artisan/{utils.slugify(artisan_slug)}/recipe"
        uri += f"/{utils.slugify(recipe_slug)}"
        params = {"locale": utils.localize(locale)}

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_proctected_resource(uri, "GET", params=params)

    @staticmethod
    def follower(
        client, region_tag: str, follower_slug: str, locale: Optional[str] = None
    ):
        """Returns the follower by slug

        Args:
            client (obj: oauth): OpenID/OAuth instance
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            follower_slug (str): the slug of a follower

        Returns:
            dict: the dict containing for the follower
        """
        uri = f"{utils.api_host(region_tag)}/d3/data/follower/{utils.slugify(follower_slug)}"
        params = {"locale": utils.localize(locale)}

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_proctected_resource(uri, "GET", params=params)

    @staticmethod
    def character_class(
        client, region_tag: str, class_slug: str, locale: Optional[str] = None
    ):
        """Returns a single character class by slug

        Args:
            client (obj: oauth): OpenID/OAuth instance
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            class_slug (str): the slug of a character class

        Returns:
            dict: the dict containing for the character class
        """
        uri = f"{utils.api_host(region_tag)}/d3/data/hero/{utils.slugify(class_slug)}"
        params = {"locale": utils.localize(locale)}

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_proctected_resource(uri, "GET", params=params)

    @staticmethod
    def api_skill(
        client,
        region_tag: str,
        class_slug: str,
        skill_slug: str,
        locale: Optional[str] = None,
    ):
        """Returns a single skill by the by slug for the specified character class

        Args:
            client (obj: oauth): OpenID/OAuth instance
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            class_slug (str): the slug of a character class
            skill_slug (str):

        Returns:
            dict: the dict containing for the skill
        """

        uri = f"{utils.api_host(region_tag)}/d3/data/hero/{utils.slugify(class_slug)}"
        uri += f"/skill/{utils.slugify(skill_slug)}"
        params = {"locale": utils.localize(locale)}

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_proctected_resource(uri, "GET", params=params)

    @staticmethod
    def item_type(
        client,
        region_tag: str,
        *,
        item_slug: Optional[str] = None,
        locale: Optional[str] = None,
    ):
        """Returns the index of item types, or a specific item type

        Args:
            client (obj: oauth): OpenID/OAuth instance
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            item_slug (str, optional): the slug of an item type

        Returns:
            dict: the dict containing for the item type
        """
        uri = f"{utils.api_host(region_tag)}/d3/data/item-type"

        if item_slug:
            uri += f"/{utils.slugify(item_slug)}"

        params = {"locale": utils.localize(locale)}

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_proctected_resource(uri, "GET", params=params)

    @staticmethod
    def item(client, region_tag: str, item_slug: str, locale: Optional[str] = None):
        """Returns the item by slug

        Args:
            client (obj: oauth): OpenID/OAuth instance
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            item_slug (str): the slug of the item

        Returns:
            dict: the dict containing for the item
        """
        uri = f"{utils.api_host(region_tag)}/d3/data/item/{utils.slugify(item_slug)}"

        params = {"locale": utils.localize(locale)}

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_proctected_resource(uri, "GET", params=params)

    @staticmethod
    def api_account(
        client, region_tag: str, bnet_tag: str, locale: Optional[str] = None
    ):
        """Returns the specified account profile

        Args:
            client (obj: oauth): OpenID/OAuth instance
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            bnet_tag (str): bnet tag of the user

        Returns:
            dict: the dict containing for the account
        """
        uri = f"{utils.api_host(region_tag)}/d3/profile/{bnet_tag}"

        params = {"locale": utils.localize(locale)}

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_proctected_resource(uri, "GET", params=params)

    @staticmethod
    def api_hero(
        client,
        region_tag: str,
        locale: str,
        bnet_tag: str,
        hero_id: str,
        category: Optional[str] = None,
    ):
        """Returns the follower by slug

        Args:
            client (obj: oauth): OpenID/OAuth instance
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            bnet_tag (str): BNet tag for the account
            hero_id (str):  Hero's ID
            category (str): category to retrieve if specified ('items', 'follower-items')

        Returns:
            dict: the dict containing for the hero, items, or follower's items
        """
        uri = (
            f"{utils.api_host(region_tag)}/d3/profile/{quote(bnet_tag)}/hero/{hero_id}"
        )

        if category:

            if category not in ("items", "follower-items"):
                raise ValueError(
                    "Invalid category;  Valid categories are 'items' and 'follower-items'"
                )

            uri += f"/{category.lower()}"

        params = {"locale": utils.localize(locale)}

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_proctected_resource(uri, "GET", params=params)



"""Generates the URI/querystring and headers for the Hearthstone API endpoints

Disclaimer:
    All rights reserved, Blizzard is the intellectual property owner of HearthstoneI and any data
    retrieved from this API.
"""

from typing import Optional, Any, Dict

from battlenet_client import utils


class Hearthstone:
    @staticmethod
    def card_search(
        client,
        region_tag: str,
        field_values: Dict[str, Any],
        locale: Optional[str] = None,
    ):
        """Searches for cards that match `field_values`

        Args:
            client (obj: oauth): OpenID/OAuth instance
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            field_values (dict): search criteria, as key/value pairs
                For more information for the field names and options:
                https://develop.battle.net/documentation/hearthstone/game-data-apis

        Returns:
            dict: json decoded search results that match `field_values`

        """
        if "gameMode" not in field_values.keys():
            field_values["gameMode"] = "constructed"

        uri = f"{utils.api_host(region_tag)}/hearthstone/cards"

        #  adding locale and namespace key/values pairs to field_values to make a complete params list
        field_values.update({"locale": utils.localize(locale)})

        try:
            return client.get(uri, params=field_values)
        except AttributeError:
            return client.fetch_proctected_resource(uri, "GET", params=field_values)

    @staticmethod
    def card(
        client,
        region_tag: str,
        card_id: str,
        *,
        locale: Optional[str] = None,
        game_mode: Optional[str] = "constructed",
    ):
        """Returns the card provided by `card_id`

        Args:
            client (obj: oauth): OpenID/OAuth instance
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            card_id (int, str): the ID or full slug of the card
            game_mode (str, optional): the game mode
                See for more information:
                https://develop.battle.net/documentation/hearthstone/guides/game-modes

        Returns:
            dict: json decoded data for the index/individual azerite essence(s)
        """
        uri = f"{utils.api_host(region_tag)}/hearthstone/cards/{card_id}"
        params = {"locale": utils.localize(locale), "gameMode": game_mode}
        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_proctected_resource(uri, params=params)

    @staticmethod
    def card_back_search(
        client,
        region_tag: str,
        field_values: Dict[str, Any],
        locale: Optional[str] = None,
    ):
        """Searches for cards that match `field_values`

        Args:
            client (obj: oauth): OpenID/OAuth instance
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            field_values (dict): search criteria, as key/value pairs
                For more information for the field names and options:
                https://develop.battle.net/documentation/hearthstone/guides/card-backs

        Returns:
            dict: json decoded search results that match `field_values`
        """
        uri = f"{utils.api_host(region_tag)}/hearthstone/cardbacks"

        #  adding locale and namespace key/values pairs to field_values to make a complete params list
        field_values.update({"locale": utils.localize(locale)})

        try:
            return client.get(uri, params=field_values)
        except AttributeError:
            return client.fetch_proctected_resource(uri, "GET", params=field_values)

    @staticmethod
    def card_back(
        client, region_tag: str, card_back_id: str, locale: Optional[str] = None
    ):
        """Returns a card back identified by `card_back_id`

        Args:
            client (obj: oauth): OpenID/OAuth instance
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            card_back_id (int, str): the ID or full slug of the card back

        Returns:
            dict: json decoded data for the card back identified by `card_back_id`
        """
        uri = f"{utils.api_host(region_tag)}/hearthstone/cards/{card_back_id}"

        params = {"locale": utils.localize(locale)}

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_proctected_resource(uri, "GET", params=params)

    @staticmethod
    def card_deck(
        client,
        region_tag: str,
        field_values: Dict[str, Any],
        locale: Optional[str] = None,
    ):
        """Searches for cards that match `field_values`

        Args:
            client (obj: oauth): OpenID/OAuth instance
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            field_values (dict): search criteria, as key/value pairs
                For more information for the field names and options:
                https://develop.battle.net/documentation/hearthstone/guides/decks

        Returns:
            dict: json decoded search results that match `field_values`
        """
        uri = f"{utils.api_host(region_tag)}/hearthstone/deck"

        #  adding locale and namespace key/values pairs to field_values to make a complete params list
        field_values.update({"locale": utils.localize(locale)})

        try:
            return client.get(uri, params=field_values)
        except AttributeError:
            return client.fetch_proctected_resource(uri, "GET", params=field_values)

    @staticmethod
    def metadata(
        client,
        region_tag: str,
        *,
        meta_data: Optional[str] = None,
        locale: Optional[str] = None,
    ):
        """Returns a list of metadata or a specific set of metadata

        Args:
            client (obj: oauth): OpenID/OAuth instance
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            meta_data (str, optional): what metadata to filter
                Please see below for more information
                https://develop.battle.net/documentation/hearthstone/guides/metadata
                valid options: `sets`, `setGroups`, `types`, `rarities`, `classes`,
                `minionTypes`, `keywords`

        Returns:
            dict: json decoded list of metadata or a specific set of metadata
        """
        uri = f"{utils.api_host(region_tag)}/hearthstone/metadata"

        if meta_data:
            uri += f"/{meta_data}"

        params = {"locale": utils.localize(locale)}

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_proctected_resource(uri, "GET", params=params)



def namespace(api_type, release, region):
    """Returns the namespace required by the WoW API endpoint"""

    if release.lower() != "retail":
        return f"{api_type}-{release.lower()}-{region.lower()}"

    return f"{api_type}-{region.lower()}"



"""Defines exceptions related to the World of Warcraft

Classes:
    WoWError
    WoWClientError
    WoWRegionError

Disclaimer:
    All rights reserved, Blizzard is the intellectual property owner of World of Warcraft and any data
    retrieved from this API.
"""


class WoWError(Exception):
    """Base Exception for the WOW API Wrappers"""

    pass


class WoWClientError(WoWError):
    """Exception used when using the wrong client, ie using Hearthstone's
    client with the WoW APIs, or when using the client credential workflow
    instead of the authorization workflow for certain APIs"""

    pass


class WoWReleaseError(WoWError):
    """Exception used when the release does not work with an API"""

    pass



from battlenet_client.constants import Region


class WoWRegion(Region):

    class Id:
        """Defines the Region IDs for World of Warcraft"""

        #: Region ID for North America
        US = 1

        #: Region ID for Taiwan
        TW = 4

        #: Region ID for Europe
        EU = 3

        #: Region ID for Korea
        KR = 2

        #: Region ID for China
        CN = 5

    class Release:
        """Defines the Release Names for World of Warcraft/World of Warcraft Classic"""

        #: Release name for the original World of Warcraft (v 1.0)
        VANILLA = 'classic1x'

        #: Release name for The Burning Crusade expansion (v 2.0)
        BURNING_CRUSADE = 'classic'

        #: Release name for the current expansion
        RETAIL = 'retail'



"""Defines the classes that handle the community APIs for Starcraft 2

Classes:
    Community
    CommunityCN

Disclaimer:
    All rights reserved, Blizzard is the intellectual property owner of Starcraft 2 and any data
    retrieved from this API.
"""

from typing import Optional

from battlenet_client import utils
from .exceptions import SC2RegionError


class Community:
    @staticmethod
    def static(
        client,
        region_tag: str,
        *,
        region_id: Optional[str] = None,
        locale: Optional[str] = None,
    ):
        """Returns all static SC2 profile data (achievements, categories, criteria, and rewards).

        Args:
            client (obj: oauth): OpenID/OAuth instance
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            region_id (str): region for the profile, or use sc2.constants  (used outside CN only)

        Returns:
            dict: dict containing the static profile data
        """
        uri = f"{utils.api_host(region_tag)}/sc2/static/profile/{region_id}"
        params = {"locale": utils.localize(locale)}
        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def metadata(
        client,
        region_tag: str,
        region_id: int,
        realm_id: int,
        profile_id: int,
        locale: Optional[str] = None,
    ):
        """Returns metadata for an individual's profile.

        Args:
            client (obj: oauth): OpenID/OAuth instance
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            region_id (int): region for the profile, or use sc2.constants
            realm_id (int): the realm of the profile (1 or 2)
            profile_id (int): the profile ID

        Returns:
            dict: dict containing the requested metadata
        """
        uri = f"{utils.api_host(region_tag)}/sc2/metadata/profile/{region_id}/{realm_id}/{profile_id}"
        params = {"locale": utils.localize(locale)}
        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def profile(
        client,
        region_tag: str,
        region_id: int,
        realm_id: int,
        profile_id: int,
        locale: Optional[str] = None,
    ):
        """Returns data about an individual SC2 profile.

        Args:
            client (obj: oauth): OpenID/OAuth instance
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            region_id (int): region for the profile, or use sc2.constants
            realm_id (int): the realm of the profile (1 or 2)
            profile_id (int): the profile ID

        Returns:
            dict: dict containing the requested profile data
        """
        uri = f"{utils.api_host(region_tag)}/sc2/profile/{region_id}/{realm_id}/{profile_id}"
        params = {"locale": utils.localize(locale)}
        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def ladder(
        client,
        region_tag: str,
        region_id: int,
        realm_id: int,
        profile_id: int,
        *,
        ladder_id: Optional[int] = "summary",
        locale: Optional[str] = None,
    ):
        """Returns a ladder summary, or specific ladder for an individual SC2 profile.

        Args:
            client (obj: oauth): OpenID/OAuth instance
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            region_id (int): region for the profile, or use sc2.constants
            realm_id (int): the realm of the profile (1 or 2)
            profile_id (int): the profile ID
            ladder_id (int, optional): the ID of a specific ladder, if desired

        Returns:
            dict: dict containing the requested profile data
        """
        uri = f"{utils.api_host(region_tag)}/sc2/profile/{region_id}/{realm_id}/{profile_id}/ladder/{ladder_id}"
        params = {"locale": utils.localize(locale)}
        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def grandmaster(
        client, region_tag: str, region_id: int, locale: Optional[str] = None
    ):
        """Returns ladder data for the current season's grandmaster leaderboard.

        Args:
            client (obj: oauth): OpenID/OAuth instance
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            region_id (int): region for the profile, or use sc2.constants

        Returns:
            dict: dict containing info about the grandmaster leaderboard
        """
        uri = f"{utils.api_host(region_tag)}/sc2/ladder/grandmaster/{region_id}"
        params = {"locale": utils.localize(locale)}
        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def season(client, region_tag: str, region_id: int, locale: Optional[str] = None):
        """Returns data about the current season.

        Args:
            client (obj: oauth): OpenID/OAuth instance
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            region_id (int): region for the profile, or use sc2.constants

        Returns:
            dict: dict containing info about the current season
        """
        uri = f"{utils.api_host(region_tag)}/sc2/ladder/season/{region_id}"
        params = {"locale": utils.localize(locale)}
        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def player(client, region_tag: str, account_id: str, locale: Optional[str] = None):
        """Returns the player data for the provided `account_id`.

        Args:
            client (obj: oauth): OpenID/OAuth instance
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            account_id (int): the account ID to request

        Returns:
            dict: json decoded data of the account
        """
        uri = f"{utils.api_host(region_tag)}/sc2/player/{account_id}"
        params = {"locale": utils.localize(locale)}
        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)


class Legacy:
    @staticmethod
    def profile(
        client,
        region_tag: str,
        region_id: int,
        realm_id: int,
        profile_id: int,
        locale: Optional[str] = None,
    ):
        """Retrieves data about an individual SC2 profile.

        Args:
            client (obj: oauth): OpenID/OAuth instance
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            region_id (int): region for the profile, or use sc2.constants
            realm_id (int): the realm of the profile (1 or 2)
            profile_id (int): the profile ID

        Returns:
             dict: json decoded data of the profile
        """
        uri = f"{utils.api_host(region_tag)}/sc2/legacy/profile/{region_id}/{realm_id}/{profile_id}"
        params = {"locale": utils.localize(locale)}
        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def ladders(
        client,
        region_tag: str,
        region_id: int,
        realm_id: int,
        profile_id: int,
        locale: Optional[str] = None,
    ):
        """Retrieves data about an individual SC2 profile's ladders.

        Args:
            client (obj: oauth): OpenID/OAuth instance
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            region_id (int): region for the profile, or use sc2.constants
            realm_id (int): the realm of the profile (1 or 2)
            profile_id (int): the profile ID

        Returns:
            dict: json decoded data of the profile's ladders
        """
        uri = f"{utils.api_host(region_tag)}/sc2/legacy/profile/{region_id}/{realm_id}/{profile_id}/ladders"
        params = {"locale": utils.localize(locale)}
        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def match_history(
        client,
        region_tag: str,
        region_id: int,
        realm_id: int,
        profile_id: int,
        locale: Optional[str] = None,
    ):
        """Returns data about an individual SC2 profile's match history.

        Args:
            client (obj: oauth): OpenID/OAuth instance
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            region_id (int): region for the profile, or use sc2.constants
            realm_id (int): the realm of the profile (1 or 2)
            profile_id (int): the profile ID

        Returns:
            dict: json decoded data of the profile's match history
        """
        uri = f"{utils.api_host(region_tag)}/sc2/legacy/profile/{region_id}/{realm_id}/{profile_id}/matches"
        params = {"locale": utils.localize(locale)}
        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def ladder(
        client,
        region_tag: str,
        region_id: int,
        ladder_id: int,
        locale: Optional[str] = None,
    ):
        """Returns data about an individual SC2 profile's match history.

        Args:
            client (obj: oauth): OpenID/OAuth instance
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            region_id (int): region for the profile, or use sc2.constants
            ladder_id (int): ladder ID for the request

        Returns:
            dict: json decoded data of the profile's data for the specified ladder
        """
        uri = f"{utils.api_host(region_tag)}/sc2/legacy/ladder/{region_id}/{ladder_id}"
        params = {"locale": utils.localize(locale)}
        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def achievements(
        client, region_tag: str, region_id: int, locale: Optional[str] = None
    ):
        """Returns the player data for the provided `account_id`.

        Args:
            client (obj: oauth): OpenID/OAuth instance
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            region_id (int): the account ID to request

        Returns:
            dict: json decoded data of the profile's achievements
        """
        uri = f"{utils.api_host(region_tag)}/sc2/legacy/data/achievements/{region_id}"
        params = {"locale": utils.localize(locale)}
        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def rewards(client, region_tag: str, region_id: int, locale: Optional[str] = None):
        """Returns the player data for the provided `account_id`.

        Args:
            client (obj: oauth): OpenID/OAuth instance
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            region_id (int): the account ID to request

        Returns:
            dict: json decoded data of the profile's rewards
        """
        uri = f"{utils.api_host(region_tag)}/sc2/legacy/data/rewards/{region_id}"
        params = {"locale": utils.localize(locale)}
        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)


class CommunityCN:
    @staticmethod
    def profile(
        client,
        region_tag: str,
        profile_id: str,
        region: str,
        name: str,
        locale: Optional[str] = None,
    ):
        """Retrieves data about an individual SC2 profile.

        Notes:
            This can only be used in the CN region

        Args:
            client (obj: oauth): OpenID/OAuth instance
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            profile_id (int): the profile ID
            region (str): region for the profile
            name (str): name of the profile

        Returns:
             dict: json decoded data of the profile
        """
        if region_tag.lower() != "cn":
            raise SC2RegionError("This API is not available in this region")

        uri = f"{utils.api_host(region_tag)}/sc2/profile/{profile_id}/{region}/{name}"
        params = {"locale": utils.localize(locale)}
        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def ladders(
        client,
        region_tag: str,
        profile_id: str,
        region: str,
        name: str,
        locale: Optional[str] = None,
    ):
        """Returns data about an individual SC2 profile's ladders.

        Args:
            client (obj: oauth): OpenID/OAuth instance
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            profile_id (int): the profile ID
            region (str): region for the profile
            name (str): name of the profile

        Returns:
            dict: json decoded data of the profile's ladders
        """

        if region_tag.lower() != "cn":
            raise SC2RegionError("This API is not available in this region")

        uri = f"{utils.api_host(region_tag)}/sc2/profile/{profile_id}/{region}/{name}/ladders"
        params = {"locale": utils.localize(locale)}
        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def match_history(
        client,
        region_tag: str,
        profile_id: str,
        region: str,
        name: str,
        locale: Optional[str] = None,
    ):
        """Returns data about an individual SC2 profile's match history.

        Args:
            client (obj: oauth): OpenID/OAuth instance
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            profile_id (int): the profile ID
            region (str): region for the profile
            name (str): name of the profile

        Returns:
            dict: json decoded data of the profile's ladders
        """
        if region_tag.lower() != "cn":
            raise SC2RegionError("This API is not available in this region")

        uri = f"{utils.api_host(region_tag)}/sc2/profile/{profile_id}/{region}/{name}/matches"
        params = {"locale": utils.localize(locale)}
        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def ladder(client, region_tag: str, ladder_id: str, locale: Optional[str] = None):
        """Returns data about an SC2 ladder.

        Args:
            client (obj: oauth): OpenID/OAuth instance
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            ladder_id (str): The ID of the ladder to retrieve.

        Returns:
            dict: dict with data of the ladder
        """
        if region_tag.lower() != "cn":
            raise SC2RegionError("This API is not available in this region")

        uri = f"{utils.api_host(region_tag)}/sc2/ladder/{ladder_id}"
        params = {"locale": utils.localize(locale)}
        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def achievements(client, region_tag: str, locale: Optional[str] = None):
        """Returns the achievements for Starcraft II

        Args:
            client (obj: oauth): OpenID/OAuth instance
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request

        Returns:
            dict: json decoded data of the guild's achievement summary
        """
        if region_tag.lower() != "cn":
            raise SC2RegionError("This API is not available in this region")

        uri = f"{utils.api_host(region_tag)}/sc2/data/achievements"
        params = {"locale": utils.localize(locale)}
        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)

    @staticmethod
    def rewards(client, region_tag: str, locale: Optional[str] = None):
        """Returns the rewards of the achievements in Starcraft II

        Args:
            client (obj: oauth): OpenID/OAuth instance
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request

        Returns:
            dict: json decoded data of the guild's achievement summary
        """
        if region_tag.lower() != "cn":
            raise SC2RegionError("This API is not available in this region")

        uri = f"{utils.api_host(region_tag)}/sc2/data/rewards"
        params = {"locale": utils.localize(locale)}
        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_protected_resource(uri, "GET", params=params)



"""Defines the methods for accessing user data from Battle.net

"""

from typing import Optional

from battlenet_client import utils


class OAuth:
    @staticmethod
    def user_info(client, region_tag: str, locale: Optional[str] = None):
        """Returns the user info

        Args:
            client (obj: oauth): OpenID/OAuth instance
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request

        Returns:
            dict: User Information (user # and battle tag ID)

        Notes:
            This function requires an OpenID, or OAuth V2 Client using the authorization code flow
        """

        url = f"{utils.auth_host(region_tag)}/oauth/userinfo"

        params = {"locale": utils.localize(locale)}

        try:
            return client.post(url, params=params)
        except AttributeError:
            return client.fetch_protected_resource(url, "POST", params=params)



from battlenet_client.constants import Region


class SC2Region(Region):

    class Id:
        """Defines the Region IDs for Starcraft II"""

        #: Region ID for North America
        US = 1

        #: Region ID for Europe
        EU = 2

        #: Region ID for Korea and Taiwan
        APAC = 3

        #: Region ID for China
        CN = 5



"""Defines exceptions related to the Diablo III API wrappers

Disclaimer:
    All rights reserved, Blizzard is the intellectual property owner of Diablo III and any data
    retrieved from this API.
"""


class HSError(Exception):
    """Base Exception for the HS API Wrappers"""

    pass


class HSClientError(HSError):
    """Exception used when using the wrong client, ie using Hearthstone's
    client with the HS APIs, or when using the client credential workflow
    instead of the authorization workflow for certain APIs"""

    pass


class HSReleaseError(HSError):
    """Exception used when the release does not work with an API"""

    pass



"""Generates the URI/querystring and headers for the Diablo III API endpoints"""
from typing import Optional

from battlenet_client import utils


class GameData:
    @staticmethod
    def season(
        client,
        region_tag: str,
        season_id: Optional[int] = None,
        locale: Optional[str] = None,
    ):
        """Returns an index of seasons, or a leaderboard of the specified season

        Args:
            client (obj: oauth): OpenID/OAuth instance
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            season_id (int): the ID of the season

        Returns:
            dict: The dict containing for the index of seasons, or the index of
                the leaderboards for the given season
        """
        uri = f"{utils.api_host(region_tag)}/data/d3/season/"
        if season_id:
            uri += season_id

        params = {"locale": utils.localize(locale)}

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_proctected_resource(uri, "GET", params=params)

    @staticmethod
    def season_leaderboard(
        client,
        region_tag: str,
        season_id: int,
        leaderboard_id: str,
        locale: Optional[str] = None,
    ):
        """Returns the leaderboard for the specified season by slug

        Args:
            client (obj: oauth): OpenID/OAuth instance
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            season_id (int): the ID of the season
            leaderboard_id (Str): the slug of the leaderboard

        Returns:
            dict: the dict containing for the leaderboard for the given season
        """
        uri = f"{utils.api_host(region_tag)}/data/d3/season/{season_id}/leaderboard/{leaderboard_id}"
        params = {"locale": utils.localize(locale)}

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_proctected_resource(uri, "GET", params=params)

    @staticmethod
    def era(
        client,
        region_tag: str,
        era_id: Optional[int] = None,
        locale: Optional[str] = None,
    ):
        """Returns an index of eras, or index of leaderboards for the era

        Args:
            client (obj: oauth): OpenID/OAuth instance
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            era_id (int): the ID of the era

        Returns:
            dict: the dict containing for the index of sea, or the leaderboard
                for the given season
        """
        uri = f"{utils.api_host(region_tag)}/data/d3/era/"
        if era_id:
            uri += era_id

        params = {"locale": utils.localize(locale)}

        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_proctected_resource(uri, "GET", params=params)

    @staticmethod
    def era_leaderboard(
        client,
        region_tag: str,
        era_id: int,
        leaderboard_id: str,
        locale: Optional[str] = None,
    ):
        """Returns the leaderboard for the specified era by slug

        Args:
            client (obj: oauth): OpenID/OAuth instance
            region_tag (str): region_tag abbreviation
            locale (str): which locale to use for the request
            era_id (int): the ID of the season
            leaderboard_id (str): the slug of the leaderboard

        Returns:
            dict: the dict containing for the leaderboard for the given season
        """
        uri = f"{utils.api_host(region_tag)}/data/d3/era/{era_id}/{leaderboard_id}"
        params = {"locale": utils.localize(locale)}
        try:
            return client.get(uri, params=params)
        except AttributeError:
            return client.fetch_proctected_resource(uri, "GET", params=params)
